package gssim.db.config;

import gssim.db.utils.DbCredentialsManager;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.security.InvalidKeyException;
import java.util.List;
import java.util.Properties;

import javax.crypto.BadPaddingException;
import javax.crypto.IllegalBlockSizeException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * The Class Config.
 */
public abstract class Config {

	protected Log log = LogFactory.getLog(Config.class);
	
	/** The stats file. */
	protected String statsFile = "Stats_Simulation 1_Simulation.txt";
	
	/** The jobs file. */
	protected String jobsFile = "Stats_Simulation 1_Jobs.txt";
	
	/** The tasks file. */
	protected String tasksFile = "Stats_Simulation 1_Raw_Tasks.txt";
	
	/** The resources file. */
	protected String resourcesFile = "Stats_Simulation 1_Raw_Resources.txt";
	
	/** The energy file. */
	protected String energyFile = "Stats_Simulation 1_Energy.txt";
	
	/** The resource load file. */
	protected String resourceLoadFile = "Stats_Simulation 1_Resource_Load.txt";
	
	/** The accumulated resources file. */
	protected String accumulatedResourcesFile = "Stats_Simulation 1_Accumulated_Resources.txt";
	
	/** The network file. */
	protected String networkFile = "Stats_Simulation 1_Network.txt";

	/** The properties file. */
	protected String propertiesFile = "experiment.properties";
	
	/** The user. */
	protected String user = "root";
	
	/** The password. */
	protected String password = null;
	
	/** The db address and name. */
	protected String db = "localhost:3306/gssim_results";
	
	/** The extract directory path (null means, that archive should be extracted in memory). */
	protected String extractDirectoryPath = null;
	
	/** Keys from properties file that should be ingored and not added to database text_property table. */
	protected List<String> propertiesIgnoreKeys;


	public Config() {}
	
	/** Constructor to fill db data from properties file */
	public Config(String propertiesPath) throws FileNotFoundException, IOException {
		Properties properties = new Properties();
		properties.load(new FileInputStream(propertiesPath));
		user = properties.getProperty("user");
		password = properties.getProperty("password");
		db = properties.getProperty("db");
	}
	
	/** Constructor to fill db data from encrypted properties file */
	public Config(String encryptedPropertiesPath, String keyPath) throws FileNotFoundException, IOException {
		
		try {
			Properties properties = DbCredentialsManager.loadPropertiesFromEncryptedFile(encryptedPropertiesPath, keyPath);
			user = properties.getProperty("user");
			password = properties.getProperty("password");
			db = properties.getProperty("db");
		} catch (InvalidKeyException e) {
			log.error("Invalid key");
		} catch (IllegalBlockSizeException e) {
			log.error("Illegal block size");
		} catch (BadPaddingException e) {
			log.error("Bad padding during decryption");
		}
	}	
	
	/**
	 * Gets the stats file.
	 *
	 * @return the stats file
	 */
	public String getStatsFile() {
		return statsFile;
	}
	
	/**
	 * Gets the jobs file.
	 *
	 * @return the jobs file
	 */
	public String getJobsFile() {
		return jobsFile;
	}

	/**
	 * Gets the tasks file.
	 *
	 * @return the tasks file
	 */
	public String getTasksFile() {
		return tasksFile;
	}

	/**
	 * Gets the resources file.
	 *
	 * @return the resources file
	 */
	public String getResourcesFile() {
		return resourcesFile;
	}

	/**
	 * Gets the energy file.
	 *
	 * @return the energy file
	 */
	public String getEnergyFile() {
		return energyFile;
	}

	/**
	 * Gets the resource load file.
	 *
	 * @return the resource load file
	 */
	public String getResourceLoadFile() {
		return resourceLoadFile;
	}

	/**
	 * Gets the accumulated resources file.
	 *
	 * @return the accumulated resources file
	 */
	public String getAccumulatedResourcesFile() {
		return accumulatedResourcesFile;
	}

	/**
	 * Gets the network file.
	 *
	 * @return the network file
	 */
	public String getNetworkFile() {
		return networkFile;
	}

	/**
	 * Gets the properties file.
	 *
	 * @return the properties file
	 */
	public String getPropertiesFile() {
		return propertiesFile;
	}

	/**
	 * Gets the user.
	 *
	 * @return the user
	 */
	public String getUser() {
		return user;
	}

	/**
	 * Gets the password.
	 *
	 * @return the password
	 */
	public String getPassword() {
		return password;
	}

	/**
	 * Gets the db.
	 *
	 * @return the db
	 */
	public String getDb() {
		return db;
	}

	/**
	 * Checks if is ignored key.
	 *
	 * @param key the key
	 * @return true, if is ignored key
	 */
	public boolean isIgnoredKey(String key) {
		return propertiesIgnoreKeys.contains(key);
	}
	
	/**
	 * Gets the extract directory path.
	 *
	 * @return the extract directory path
	 */
	public String getExtractDirectoryPath() {
		return extractDirectoryPath;
	}
}
