package gssim.gssimstudio.editor.gssim.topology
{
	import flash.events.Event;
	import flash.events.EventDispatcher;
	
	/**
	 *  Dispatched when there is any change to the nodes and/or links of this graph.
	 *
	 *  @eventType flash.events.Event
	 */
	[Event(name="changed", type="flash.events.Event")]
	
	/**
	 *  A Graph is a collection of items that can be linked to each other.
	 */
 	public class Graph extends EventDispatcher
	{
		public static const CHANGE:String = "change";

		public function Graph(): void {
		}
		
		private var _nodes: Object = new Object(); // map of id -> GraphItem
		private var _edges: Object = new Object(); // map of id -> (map of id -> 0)
		private var nodeArray: Array/*of GraphItem*/; 
		private var edgeArray: Array/*of [GraphItem, GraphItem]*/;
		private var _distinguishedGraphItem: GraphItem;

	    /**
	     *  Removes an item from the graph.
	     *
	     *  @param item The item that you want to remove from the graph.
	     */
		public function remove(item:GraphItem):void
		{
			delete _nodes[item.id];
			delete _edges[item.id];
			
			for (var id: String in _edges) {
				var friends: Object = _edges[id];
				delete friends[item.id];
			}
			
			nodeArray = null;
			edgeArray = null;
			changed();
		}
		
	    /**
	     *  Remove the link between 2 items.
	     *
	     *  @param item1 an item in the graph that is linked to item2
	     *  @param item2 an item in the graph that is linked to item1
	     */
		public function unlink(item1:GraphItem, item2:GraphItem):void
		{
			var friends: Object = _edges[item1.id];
			delete friends[item2.id];
			
			friends = _edges[item2.id];
			delete friends[item1.id];
			
			edgeArray = null;
			changed();
		}
		
	    /**
	     *  An array of all the links in the graph.
	     *  Each array element is an array of 2 strings, 
	     *  which are the ids of two items that are linked.
	     */
		public function get edges():Array
		{
			if(edgeArray == null) {
				edgeArray = new Array();
				var done: Object = new Object();
				for (var id: String in _edges) {
					done[id] = true;
					var friends: Object = _edges[id];
					for (var friendID: String in friends) {
						if(!done.hasOwnProperty(friendID))
							edgeArray.push([_nodes[id], nodes[friendID]]);
					}
				}
			}
			return edgeArray;
		}
		
	    /**
	     *  An associative array of all the items in the graph.
	     *  The key is the id, the value is the GraphItem.
	     */
		public function get nodes():Object
		{
			return _nodes;
		}
		
	    /**
	     *  True if this graph has any nodes at all.
	     */
		public function get hasNodes(): Boolean
		{
			for each (var item: GraphItem in _nodes) {
				return true;
			}
			return false;
		}
		
	    /**
	     *  How many items are in this graph.
	     */
		public function get nodeCount(): int
		{
			var result:  int = 0;
			for each (var item: GraphItem in _nodes) {
				result++;
			}
			return result;
		}
		
	    /**
	     *  Link 2 items. This has no effect if the 2 items are already
	     *  linked. Links are not directional: link(a,b) is equivalent to
	     *  link(b,a).
	     *
	     *  @param item1 an item in the graph
	     *  @param item2 an item in the graph
	     *  @param data any data you like, or null. 
	     */
		public function link(item1:GraphItem, item2:GraphItem, data: Object = null):void
		{
			if(data == null) data = 0;
			
			var friends: Object = _edges[item1.id];
			friends[item2.id] = data;
			
			friends = _edges[item2.id];
			friends[item1.id] = data;
			
			edgeArray = null;
			changed();
		}
	
	    /**
	     *  Add an item to the graph.
	     *
	     *  @param item an item to add to the graph
	     */
		public function add(item:GraphItem):void
		{
			if(_distinguishedGraphItem == null)
				_distinguishedGraphItem = item;
				
			if(_nodes.hasOwnProperty(item.id)) {
				return;
			}
			
			_nodes[item.id] = item;
			_edges[item.id] = new Object();
			changed();
		}
		
	    /**
	     *  Find out if two items are linked.
	     *
	     *  @param item1 an item in the graph
	     *  @param item2 an item in the graph
	     *
	     *  @return true if the two items are linked to each other.
	     */
		public function linked(item1:GraphItem, item2:GraphItem):Boolean
		{
			var friends: Object = _edges[item1.id];
			return (friends != null) && friends.hasOwnProperty(item2.id);
		}
		
	    /**
	     *  retrieve the data that is associated with a link.
	     *
	     *  @param item1 an item in the graph
	     *  @param item2 an item in the graph
	     *
	     *  @return Object the data that was associated with the link between the two items.
	     *  If no data, or null, was associated with the link, we return 0. If there is no link
	     *  between the items, we return null.
	     */
		public function getLinkData(item1:GraphItem, item2:GraphItem):Object
		{
			var friends: Object = _edges[item1.id];
			
			if ((friends != null) && friends.hasOwnProperty(item2.id))
				return friends[item2.id];
			else
				return null;
		}
		
	    /**
	     *  Find out how many items are linked to a given item.
	     *
	     *  @param item an item in the graph
	     *
	     *  @return thes number of items to which this item is linked.
	     */
		public function numLinks(item: GraphItem): int {
			var friends: Object = _edges[item.id];
			var result: int = 0;
			for (var i: String in friends) { 
				result++; 
			}
			return result;
		}
		
	    /**
	     *  Find out if an item with a given id exists in the graph.
	     *
	     *  @param id any String
	     *
	     *  @return true if there is an item in the graph with the given id,
	     *  false otherwise.
	     */
		public function hasNode(id: String): Boolean {
			return _nodes.hasOwnProperty(id);
		}
		
	    /**
	     *  Find an item in the graph by id.
	     *
	     *  @param id any String
	     *
	     *  @return the item in the graph that has the given id,
	     *  or null if there is no such item.
	     */
		public function find(id: String): GraphItem {
			if(_nodes.hasOwnProperty(id))
				return _nodes[id];
			return null;
		}
		
	    /**
	     *  Get an array of all the items that a given item is linked to.
	     *
	     *  @param id any String
	     *
	     *  @return an array of GraphItems
	     */
		public function neighbors(id: String): Object {
			return _edges[id];
		}
		
	    /** Sometimes it's handy for the graph to remember one particular item.
	     * You can use this for any purpose you like, it's not used internally by the Graph.
	     * By default, the distinguished item is the first item that was added to this graph.
	     */
		public function get distinguishedGraphItem(): GraphItem {
			return _distinguishedGraphItem;
		}
		
		public function set distinguishedGraphItem(item: GraphItem): void {
			_distinguishedGraphItem = item;
		}
		
		/** Remove all items from the graph. */
		public function empty(): void {
			_nodes = new Object();
			_edges = new Object();
			nodeArray = null;
			edgeArray = null;
			_distinguishedGraphItem = null;
			changed();	
		}
		
		private function changed(): void {
			dispatchEvent(new Event(CHANGE));
		}
		
		/** Checks if graph is strongly connected. */
		public function isStronglyConnected() : Boolean
		{
			if (!hasNodes) return false;
			var id : String = distinguishedGraphItem.id;
			for each (var node : GraphItem in nodes)
			{
				if (node is UserObject) id = node.id;
				node.visited = false;
			}
			dfs(find(id));
			for each (var node2 : GraphItem in nodes)
			{
				if (!node2.visited) return false;
			}
			return true;
		}
		
		/** Depth-first-search */
		private function dfs(item : GraphItem) : void
		{
			item.visited = true;
			for (var id : String in neighbors(item.id))
			{
				var node : GraphItem = find(id); 
				if (!node.visited) dfs(node);
			}
		}
		
		public override function toString():String
		{
			var result : String = "";
			var countRouters : Number = 0;
			var routers : Object = new Object(); //map id -> router
			var countLinks : Number = 0;
			var links : Object = new Object(); // map id -> link
			var countLinksRes : Number = 0;
			var links_res : Object = new Object(); // map id -> link
			var countLinksUser : Number = 0;
			var links_user : Object = new Object(); // map id -> link
			for each (var node : GenderObject in _nodes)
			{
				if (node is RouterObject)
				{
					routers[node.id] = node;
					countRouters++;
				}
				for each(var link : NetworkObjectLinker in node.links)
				{
					if (link.NodeFrom is RouterObject)
					{
						if (link.NodeTo is RouterObject)
						{
							if (!links.hasOwnProperty(link.id))
							{
								links[link.id] = link;
								countLinks++;
							}
						}
						else if (link.NodeTo is ResourceObject)
						{
							if (!links_res.hasOwnProperty(link.id))
							{
								links_res[link.id] = link;
								countLinksRes++;
							}
						}
						else if (link.NodeTo is UserObject)
						{
							if (!links_user.hasOwnProperty(link.id))
							{
								links_user[link.id] = link;
								countLinksUser++;
							}
						}						
					}
					else if (link.NodeFrom is ResourceObject)
					{
						if (!links_res.hasOwnProperty(link.id))
						{
							links_res[link.id] = link;
							countLinksRes++;
						}
					}
					else if (link.NodeFrom is UserObject)
					{
						if (!links_user.hasOwnProperty(link.id))
						{
							links_user[link.id] = link;
							countLinksUser++;
						}
					}	
				}
			}
			//build final string
			result += countRouters.toString() + "\r\n";
			for each(var router : RouterObject in routers)
			{
				result += router.nodeName + "\r\n";
			}
			result += countLinks.toString() + "\r\n";
			for each(var _link : NetworkObjectLinker in links)
			{
				result += _link.NodeFrom.nodeName + " " + _link.NodeTo.nodeName + " ";
				result += _link.BaudRate.toString() + " ";
				result += _link.PropDelay.toString() + " ";
				result += _link.MTU.toString() + "\r\n";
			}
			result += countLinksRes.toString() + "\r\n";
			for each(var _link_res : NetworkObjectLinker in links_res)
			{
				if (_link_res.NodeFrom is ResourceObject)
					result += _link_res.NodeFrom.nodeName + " " + _link_res.NodeTo.nodeName + " ";
				else
					result += _link_res.NodeTo.nodeName + " " + _link_res.NodeFrom.nodeName + " ";
				result += _link_res.BaudRate.toString() + " ";
				result += _link_res.PropDelay.toString() + " ";
				result += _link_res.MTU.toString() + "\r\n";
			}
			result += countLinksUser.toString() + "\r\n";
			for each(var _link_user : NetworkObjectLinker in links_user)
			{
				if (_link_user.NodeFrom is UserObject)
					result += _link_user.NodeFrom.nodeName + " " + _link_user.NodeTo.nodeName + " ";
				else
					result += _link_user.NodeTo.nodeName + " " + _link_user.NodeFrom.nodeName + " ";
				result += _link_user.BaudRate.toString() + " ";
				result += _link_user.PropDelay.toString() + " ";
				result += _link_user.MTU.toString() + "\r\n";
			}
			result += nodeCount.toString() + "\r\n";
			for each (var node2 : GenderObject in _nodes)
			{
				result += node2.nodeName + " ";
				result += node2.x.toString() + " ";
				result += node2.y.toString() + "\r\n";
			}
			links = null;
			links_res = null;
			links_user = null;
			routers = null;
			return result;
		}
		
		 /**
	     *  Translate whole graph by vector (tx, ty)
	     *
	     *  @param tx translate along X-axis
	     *  @param ty translate along Y-axis
	     */
		public function translate(tx : Number, ty : Number) : void
		{
			for each (var node : GenderObject in _nodes)
			{
				node.x += tx;
				node.y += ty;
				node.dispatchMoveAction();
			}
		} 
	}
}
